import 'package:flutter/material.dart';

import '../form.dart';

/// A single [FastChoiceChips] chip.
@immutable
class FastChoiceChip
    implements
        ChipAttributes,
        SelectableChipAttributes,
        DisabledChipAttributes {
  FastChoiceChip({
    Widget? label,
    this.autofocus = false,
    this.avatar,
    this.avatarBorder = const CircleBorder(),
    this.backgroundColor,
    this.checkmarkColor,
    this.clipBehavior = Clip.none,
    this.color,
    this.disabledColor,
    this.elevation,
    this.enabled = true,
    this.focusNode,
    this.iconTheme,
    this.labelStyle,
    this.labelPadding,
    this.materialTapTargetSize,
    this.onSelected,
    this.padding,
    this.pressElevation,
    this.selected = false,
    this.selectedColor,
    this.selectedShadowColor,
    this.shadowColor,
    this.shape,
    this.showCheckmark,
    this.side,
    this.surfaceTintColor,
    this.tooltip,
    required this.value,
    this.visualDensity,
  }) : label = label is Widget ? label : Text(value);

  @override
  final bool autofocus;
  @override
  final Widget? avatar;
  @override
  final ShapeBorder avatarBorder;
  @override
  final Color? backgroundColor;
  @override
  final Clip clipBehavior;
  @override
  final MaterialStateProperty<Color?>? color;
  @override
  final Color? disabledColor;
  @override
  final double? elevation;
  @override
  final FocusNode? focusNode;
  @override
  final IconThemeData? iconTheme;
  @override
  final Widget label;
  @override
  final EdgeInsetsGeometry? labelPadding;
  @override
  final TextStyle? labelStyle;
  @override
  final MaterialTapTargetSize? materialTapTargetSize;
  @override
  final EdgeInsetsGeometry? padding;
  @override
  final double? pressElevation;
  @override
  final ValueChanged<bool>? onSelected;
  @override
  final bool selected;
  @override
  final Color? selectedColor;
  @override
  final Color? selectedShadowColor;
  @override
  final Color? shadowColor;
  @override
  final OutlinedBorder? shape;
  @override
  final BorderSide? side;
  @override
  final Color? surfaceTintColor;
  @override
  final String? tooltip;
  @override
  final VisualDensity? visualDensity;

  final bool enabled;
  final Color? checkmarkColor;
  final bool? showCheckmark;
  final String value;

  @override
  bool get isEnabled => enabled;
}

typedef FastChoiceChipBuilder = Widget Function(
    FastChoiceChip chip, FastChoiceChipsState field);

/// A [FastFormField] that contains a list of [ChoiceChip].
@immutable
class FastChoiceChips extends FastFormField<List<String>> {
  FastChoiceChips({
    List<String>? initialValue,
    super.autovalidateMode,
    super.builder = choiceChipsBuilder,
    super.contentPadding,
    super.decoration,
    super.enabled,
    super.helperText,
    super.key,
    super.labelText,
    required super.name,
    super.onChanged,
    super.onReset,
    super.onSaved,
    super.restorationId,
    super.validator,
    this.alignment = WrapAlignment.start,
    this.chipBuilder = choiceChipBuilder,
    this.chipPadding,
    required this.chips,
    this.clipBehavior = Clip.none,
    this.crossAlignment = WrapCrossAlignment.start,
    this.direction = Axis.horizontal,
    this.spacing = 12.0,
    this.runAlignment = WrapAlignment.start,
    this.runSpacing = 0.0,
    this.showCheckmark,
    this.textDirection,
    this.verticalDirection = VerticalDirection.down,
  }) : super(initialValue: initialValue ?? _getInitialValue(chips));

  final WrapAlignment alignment;
  final FastChoiceChipBuilder chipBuilder;
  final EdgeInsetsGeometry? chipPadding;
  final List<FastChoiceChip> chips;
  final Clip clipBehavior;
  final WrapCrossAlignment crossAlignment;
  final Axis direction;
  final WrapAlignment runAlignment;
  final double runSpacing;
  final bool? showCheckmark;
  final double spacing;
  final TextDirection? textDirection;
  final VerticalDirection verticalDirection;

  @override
  FastChoiceChipsState createState() => FastChoiceChipsState();
}

/// State associated with a [FastChoiceChips] widget.
class FastChoiceChipsState extends FastFormFieldState<List<String>> {
  @override
  FastChoiceChips get widget => super.widget as FastChoiceChips;
}

/// Fallback for setting the default [FastChoiceChips.initialValue].
///
/// Returns a [List] that contains every [FastChoiceChip.value] in
/// [FastChoiceChips.chips] where [FastChoiceChip.selected] is `true`.
List<String> _getInitialValue<T>(List<FastChoiceChip> chips) {
  return chips
      .where((chip) => chip.selected)
      .map((chip) => chip.value)
      .toList();
}

/// A [FastChoiceChipBuilder] that is the default [FastChoiceChips.chipBuilder].
///
/// Returns a [ChoiceChip] that updates the [FastChoiceChipsState.value] in
/// [ChoiceChip.onSelected].
ChoiceChip choiceChipBuilder(FastChoiceChip chip, FastChoiceChipsState field) {
  final FastChoiceChipsState(:didChange, :enabled, :value!, :widget) = field;

  void onSelected(selected) {
    chip.onSelected?.call(selected);

    List<String> updatedValue;
    if (selected) {
      updatedValue = [...value, chip.value];
    } else {
      updatedValue = [...value]..remove(chip.value);
    }

    didChange(updatedValue);
  }

  return ChoiceChip(
    autofocus: chip.autofocus,
    avatar: chip.avatar,
    avatarBorder: chip.avatarBorder,
    backgroundColor: chip.backgroundColor,
    checkmarkColor: chip.checkmarkColor,
    clipBehavior: chip.clipBehavior,
    color: chip.color,
    disabledColor: chip.disabledColor,
    elevation: chip.elevation,
    focusNode: chip.focusNode,
    iconTheme: chip.iconTheme,
    label: chip.label,
    labelPadding: chip.labelPadding,
    labelStyle: chip.labelStyle,
    materialTapTargetSize: chip.materialTapTargetSize,
    onSelected: enabled && chip.isEnabled ? onSelected : null,
    padding: chip.padding ?? widget.chipPadding,
    pressElevation: chip.pressElevation,
    selected: value.contains(chip.value),
    selectedColor: chip.selectedColor,
    selectedShadowColor: chip.selectedShadowColor,
    shadowColor: chip.shadowColor,
    shape: chip.shape,
    showCheckmark: chip.showCheckmark ?? widget.showCheckmark,
    side: chip.side,
    surfaceTintColor: chip.surfaceTintColor,
    tooltip: chip.tooltip,
    visualDensity: chip.visualDensity,
  );
}

/// A [FormFieldBuilder] that is the default [FastChoiceChips.builder].
///
/// Uses [FastChoiceChips.chipBuilder] to build a [ChoiceChip] for every
/// [FastChoiceChip] in [FastChoiceChips.chips].
///
/// Returns an [InputDecorator] that wraps a [List] of [ChoiceChip] on any
/// [TargetPlatform].
Widget choiceChipsBuilder(FormFieldState field) {
  final FastChoiceChipsState(:decoration, :widget) =
      field as FastChoiceChipsState;

  return InputDecorator(
    decoration: decoration,
    child: Wrap(
      alignment: widget.alignment,
      crossAxisAlignment: widget.crossAlignment,
      clipBehavior: widget.clipBehavior,
      direction: widget.direction,
      runAlignment: widget.runAlignment,
      runSpacing: widget.runSpacing,
      spacing: widget.spacing,
      textDirection: widget.textDirection,
      verticalDirection: widget.verticalDirection,
      children: [
        for (final chip in widget.chips) widget.chipBuilder(chip, field),
      ],
    ),
  );
}
